"""
Genesis parse
-------------
It implements the intelligence necessary for us to transform freeswitch events into python primitive types.
"""

from typing import Optional
from collections import UserDict
from urllib.parse import unquote


class ESLEvent(UserDict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.body: Optional[str] = None


def parse_headers(payload: str) -> ESLEvent:
    lines = payload.strip().splitlines()
    headers = ESLEvent()
    buffer = ""
    value = ""

    for line in lines:
        if ": " in line:
            key, value = line.split(": ", 1)
            buffer = key

        else:
            value += "\n" + line
            key = buffer

        key = unquote(key.strip(), encoding="UTF-8")
        value = unquote(value.strip(), encoding="UTF-8")

        if ": " in line and key in headers:
            backup = headers[key]

            if isinstance(backup, str):
                headers[key] = [backup, value]
            else:
                headers[key] = [*backup, value]

        else:
            headers[key] = value

    return headers



"""
Genesis inbound
---------------
ESL implementation used for incoming connections on freeswitch.
"""

from __future__ import annotations

from asyncio import open_connection, TimeoutError, wait_for
from typing import Awaitable

from genesis.exceptions import ConnectionTimeoutError, AuthenticationError
from genesis.protocol import Protocol
from genesis.logger import logger
from opentelemetry import trace, metrics

tracer = trace.get_tracer(__name__)
meter = metrics.get_meter(__name__)

active_connections_counter = meter.create_up_down_counter(
    "genesis.connections.active",
    description="Number of active connections",
    unit="1",
)
connection_errors_counter = meter.create_counter(
    "genesis.connections.errors",
    description="Number of connection errors",
    unit="1",
)


class Inbound(Protocol):
    """
    Inbound class
    -------------

    Given a valid set of information, establish a connection to a freeswitch server.

    Attributes:
    - host: required
        IP address associated with the connection destination server.
    - port: required
        Network port where ESL module is listening.
    - password: required
        Password used for authentication on freeswitch.
    - timeout: optional
        Maximum time we wait to initiate a connection.
    """

    def __init__(self, host: str, port: int, password: str, timeout: int = 5) -> None:
        super().__init__()
        self.password = password
        self.timeout = timeout
        self.host = host
        self.port = port

    async def __aenter__(self) -> Inbound:
        """Interface used to implement a context manager."""
        await self.start()
        return self

    async def __aexit__(self, *args, **kwargs) -> None:
        """Interface used to implement a context manager."""
        await self.stop()

    async def authenticate(self) -> None:
        """Authenticates to the freeswitch server. Raises an exception on failure."""
        await self.authentication_event.wait()
        logger.debug("Send command to authenticate inbound ESL connection.")
        response = await self.send(f"auth {self.password}")

        if response["Reply-Text"] != "+OK accepted":
            logger.debug("Freeswitch said the passed password is incorrect.")
            connection_errors_counter.add(1, attributes={"error": "authentication_failed", "type": "inbound"})
            raise AuthenticationError("Invalid password")

    async def start(self) -> None:
        """Initiates an authenticated connection to a freeswitch server."""
        try:
            try:
                with tracer.start_as_current_span(
                    "inbound_connect",
                    attributes={
                        "net.peer.name": self.host,
                        "net.peer.port": self.port,
                    },
                ):
                    promise = open_connection(self.host, self.port)
                    self.reader, self.writer = await wait_for(promise, self.timeout)
            except Exception as tracer_error:
                # OTel not initialized - connect without tracing
                if "tracer" not in str(tracer_error).lower():
                    raise
                promise = open_connection(self.host, self.port)
                self.reader, self.writer = await wait_for(promise, self.timeout)
        except TimeoutError:
            logger.debug("A timeout occurred when trying to connect to the freeswitch.")
            try:
                connection_errors_counter.add(1, attributes={"error": "timeout", "type": "inbound"})
            except Exception:
                pass
            raise ConnectionTimeoutError()

        await super().start()
        try:
            active_connections_counter.add(1, attributes={"type": "inbound"})
        except Exception as e:
            logger.error(f"OTel error in start: {e}")
            pass
        await self.authenticate()

    async def stop(self) -> None:
        """Terminates the connection."""
        await super().stop()
        try:
            active_connections_counter.add(-1, attributes={"type": "inbound"})
        except Exception:
            pass



"""
Consumer Module
---------------
Simple abstraction used to put some syntactic sugar into freeswitch event consumption.
"""

from typing import Callable, Optional, Coroutine, Union
import functools
import asyncio
import re

from genesis.inbound import Inbound
from genesis.logger import logger


def filtrate(
    key: str, value: Optional[str] = None, regex: bool = False
) -> Union[Callable, Coroutine]:
    """
    Method that allows to filter the events according to a set 'key', 'value'.

    Parameters
    ----------
    - key: required
        Key to be searched in the event.
    - value: optional
        Value needed in the last key.
    - regex: optional
        Tells whether 'value' is a regular expression.
    """

    def decorator(function: Union[Callable, Coroutine]):
        @functools.wraps(function)
        async def wrapper(message):
            if isinstance(message, dict):
                if key in message:
                    content = message[key]

                    if value is None:
                        result = function(message)
                        if asyncio.iscoroutine(result):
                            return await result
                        return result

                    if not regex and content == value:
                        result = function(message)
                        if asyncio.iscoroutine(result):
                            return await result
                        return result

                    if regex and re.match(value, content):
                        result = function(message)
                        if asyncio.iscoroutine(result):
                            return await result
                        return result

        return wrapper

    return decorator


class Consumer:
    """
    Consumer class
    --------------
    Abstraction used to create valid event consumers.

    Attributes:
    - host: required
        IP address associated with the connection destination server.
    - port: required
        Network port where ESL module is listening.
    - password: required
        Password used for authentication on freeswitch.
    - timeout: optional
        Maximum time we wait to initiate a connection.
    """

    def __init__(
        self,
        host: str = "127.0.0.1",
        port: int = 8021,
        password: str = "ClueCon",
        timeout: int = 5,
    ) -> None:
        self.protocol: Inbound = Inbound(host, port, password, timeout)

    def handle(self, event: str) -> Union[Callable, Coroutine]:
        """Decorator that allows the registration of new handlers.

        Parameters
        ----------
        - event: required
            Name of the event to be parsed.
        """

        def decorator(function: Union[Callable, Coroutine]):
            self.protocol.on(event, function)

            @functools.wraps(function)
            async def wrapper(*args, **kwargs):
                return await function(*args, **kwargs)

            return wrapper

        return decorator

    async def wait(self) -> None:
        while bool(self.protocol.is_connected):
            logger.debug("Wait to receive new events...")
            await asyncio.sleep(1)

    async def start(self) -> None:
        """Method called to request the freeswitch to start sending us the appropriate events."""
        try:
            async with self.protocol as protocol:
                logger.debug("Asking freeswitch to send us all events.")
                await protocol.send("events plain ALL")

                for event in protocol.handlers.keys():
                    logger.debug(
                        f"Requesting freeswitch to filter events of type '{event}'."
                    )

                    if event.isupper():
                        logger.debug(
                            f"Send command to filtrate events with name: '{event}'."
                        )
                        await protocol.send(f"filter Event-Name {event}")
                    else:
                        logger.debug(
                            f"Send command to filtrate events with subclass: '{event}'."
                        )
                        await protocol.send(f"filter Event-Subclass {event}")

                await self.wait()

        except:
            await self.stop()
            raise

    async def stop(self) -> None:
        await self.protocol.stop()



import asyncio
import fnmatch
from pathlib import Path
from typing import Optional

from watchdog.observers import Observer
from watchdog.events import FileSystemEvent, FileSystemEventHandler


class EventHandler(FileSystemEventHandler):
    """
    Event handler
    -------------

    This class is responsible for handling the events from the filesystem.
    """

    def __init__(
        self, queue: asyncio.Queue, loop: asyncio.BaseEventLoop, *args, **kwargs
    ):
        self._loop = loop
        self._queue = queue
        super(*args, **kwargs)

    def on_any_event(self, event: FileSystemEvent) -> None:
        if fnmatch.fnmatch(event.src_path, "*.py"):
            self._loop.call_soon_threadsafe(self._queue.put_nowait, event)


class EventIterator(object):
    """ "
    Event iterator
    --------------

    This class is responsible for iterating over the events from the queue.
    """

    def __init__(
        self, queue: asyncio.Queue, loop: Optional[asyncio.BaseEventLoop] = None
    ):
        self.queue = queue

    def __aiter__(self):
        return self

    async def __anext__(self):
        item = await self.queue.get()
        if item is not None:
            return item


def factory(
    path: Path,
    queue: asyncio.Queue,
    loop: asyncio.BaseEventLoop,
    recursive: bool = True,
) -> Observer:
    """This function creates an observer instance and returns it."""
    handler = EventHandler(queue, loop)

    observer = Observer()
    observer.schedule(handler, str(path), recursive=recursive)

    return observer



from typing import Annotated, Union
from pathlib import Path
import importlib
import logging
import asyncio

import typer


from genesis.cli import watcher
from genesis.logger import logger
from genesis.consumer import Consumer
from genesis.cli.exceptions import CLIExcpetion
from genesis.cli.utils import complete_log_levels
from genesis.cli.discover import get_import_string

consumer = typer.Typer(rich_markup_mode="rich")


async def _run_with_reload(app: Consumer, path: Path) -> None:
    loop = asyncio.get_running_loop()
    queue = asyncio.Queue()

    async def consume(queue: asyncio.Queue) -> None:
        await app.start()
        async for event in watcher.EventIterator(queue):
            if event:
                logger.info(f"File changed: {event.src_path}")
                await app.stop()
                logger.info("App stopped, restarting...")
                await app.start()

    observer = watcher.factory(path, queue, loop)

    observer.start()
    asyncio.run_coroutine_threadsafe(queue.put(None), loop)

    try:
        await consume(queue)
        while True:
            await asyncio.sleep(1)
    except KeyboardInterrupt:
        observer.stop()


def _run(
    path: Union[Path, None] = None,
    host: str = "127.0.0.1",
    port: int = 8021,
    reload: bool = True,
    app: Union[str, None] = None,
    loglevel: str = "info",
    password: str = "ClueCon",
) -> None:
    try:
        import_string = get_import_string(Consumer, path=path, app_name=app)

        logger.info(f"Consumer started - Host: {host}, Port: {port}")

        module_str, attr_str = import_string.split(":")
        module = importlib.import_module(module_str)
        app: Consumer = getattr(module, attr_str)

        app.host = host
        app.port = port
        app.password = password

        logger.info(f"Setting log level to [bold]{loglevel.upper()}[/bold]")
        levels = logging.getLevelNamesMapping()
        logger.setLevel(levels.get(loglevel.upper(), logging.INFO))

        if reload:
            asyncio.run(_run_with_reload(app, path))
        else:
            asyncio.run(app.start())

    except CLIExcpetion as e:
        logger.error(e)
        raise typer.Exit(1)


@consumer.command()
def dev(
    path: Annotated[
        Path,
        typer.Argument(
            help="A path to a Python file or package directory.", metavar="PATH"
        ),
    ],
    *,
    host: Annotated[
        str,
        typer.Option(help="The host to connect on.", envvar="ESL_HOST"),
    ] = "127.0.0.1",
    port: Annotated[
        int,
        typer.Option(help="The port to connect on.", envvar="ESL_PORT"),
    ] = 8021,
    password: Annotated[
        str,
        typer.Option(
            help="The password to authenticate on host.", envvar="ESL_PASSWORD"
        ),
    ] = "ClueCon",
    app: Annotated[
        Union[str, None],
        typer.Option(
            help="Variable that contains the [bold]Consumer[/bold] app in the imported module or package.",
            envvar="ESL_APP_NAME",
        ),
    ] = None,
    loglevel: Annotated[
        str,
        typer.Option(
            help="The log level to use.",
            envvar="ESL_LOG_LEVEL",
            show_default=True,
            case_sensitive=False,
            autocompletion=complete_log_levels,
        ),
    ] = "info",
):
    """
    Run a [bold]Consumer[/bold] genesis app in [yellow]development[/yellow] mode. 🧪

    It automatically detects the Python module or package that needs to be imported based on the file or directory path passed.

    It detects the [bold]Consumer[/bold] app object to use based on the app name passed.
    By default it looks in the module or package for an object named [blue]app[/blue].
    Otherwise, it uses the first [bold]Consumer[/bold] app found in the imported module or package.
    """
    _run(
        path=path,
        host=host,
        port=port,
        password=password,
        app=app,
        reload=True,
        loglevel=loglevel,
    )


@consumer.command()
def run(
    path: Annotated[
        Path,
        typer.Argument(
            help="A path to a Python file or package directory.", metavar="PATH"
        ),
    ],
    *,
    host: Annotated[
        str,
        typer.Option(help="The host to connect on.", envvar="ESL_HOST"),
    ] = "127.0.0.1",
    port: Annotated[
        int,
        typer.Option(help="The port to connect on.", envvar="ESL_PORT"),
    ] = 8021,
    password: Annotated[
        str,
        typer.Option(
            help="The password to authenticate on host.", envvar="ESL_PASSWORD"
        ),
    ] = "ClueCon",
    app: Annotated[
        Union[str, None],
        typer.Option(
            help="Variable that contains the [bold]Consumer[/bold] app in the imported module or package.",
            envvar="ESL_APP_NAME",
        ),
    ] = None,
    loglevel: Annotated[
        str,
        typer.Option(
            help="The log level to use.",
            envvar="ESL_LOG_LEVEL",
            show_default=True,
            case_sensitive=False,
            autocompletion=complete_log_levels,
        ),
    ] = "info",
):
    """
    Run a [bold]Consumer[/bold] genesis app in [green]production[/green] mode. 🚀

    It automatically detects the Python module or package that needs to be imported based on the file or directory path passed.

    It detects the [bold]Consumer[/bold] app object to use based on the app name passed.
    By default it looks in the module or package for an object named [blue]app[/blue].
    Otherwise, it uses the first [bold]Consumer[/bold] app found in the imported module or package.
    """
    _run(
        path=path,
        host=host,
        port=port,
        password=password,
        app=app,
        reload=False,
        loglevel=loglevel,
    )



"""
Genesis outbound
----------------

ESL implementation used for outgoing connections on freeswitch.
"""

from __future__ import annotations

from asyncio import StreamReader, StreamWriter, Queue, start_server, Event, wait_for
from typing import Optional, Union, Dict, Literal
from collections.abc import Callable, Coroutine
from functools import partial
from pprint import pformat
from uuid import uuid4
import socket

from genesis.protocol import Protocol
from genesis.parser import ESLEvent
from genesis.logger import logger
from opentelemetry import trace, metrics

tracer = trace.get_tracer(__name__)
meter = metrics.get_meter(__name__)

active_connections_counter = meter.create_up_down_counter(
    "genesis.connections.active",
    description="Number of active connections",
    unit="1",
)


class Session(Protocol):
    """
    Session class
    -------------

    Abstracts a session established between the application and the freeswitch.

    Attributes:
    - reader: required
        StreamReader used to read incoming information.
    - writer: required
        StreamWriter used to send information to freeswitch.
    """

    def __init__(self, reader: StreamReader, writer: StreamWriter) -> None:
        super().__init__()
        self.context: Dict[str, str] = dict()
        self.reader = reader
        self.writer = writer
        self.fifo = Queue()

    async def __aenter__(self) -> Session:
        """Interface used to implement a context manager."""
        await self.start()
        return self

    async def __aexit__(self, *args, **kwargs) -> None:
        """Interface used to implement a context manager."""
        await self.stop()

    async def _awaitable_complete_command(
        self, event_uuid: str, timeout: Optional[int] = None
    ) -> Event:
        """
        Create an event that will be set when a command completes.

        Args:
            event_uuid: UUID to track the specific command execution.
            timeout: Optional timeout in seconds for the command to complete.

        Raises:
            asyncio.TimeoutError: if the command does not complete within the timeout period.

        Returns:
            Event that will be set when command completes.
        """
        semaphore = Event()

        handlers = {}

        async def cleanup():
            for key, value in handlers.items():
                self.remove(key, value)

        async def channel_execute_complete_handler(session: Session, event: ESLEvent):
            logger.debug(f"Received channel execute complete event: {event}")

            if "Application-UUID" in event and event["Application-UUID"] == event_uuid:
                await session.fifo.put(event)
                semaphore.set()
                await cleanup()

        # Handler for CHANNEL_HANGUP_COMPLETE event to ensure we don't miss it
        # if the call is hung up before the command completes
        async def channel_hangup_complete_handler(session: Session, event: ESLEvent):
            logger.debug(f"Received hangup event: {event}")

            if (
                "Unique-ID" in event
                and session.context.get("Channel-Unique-ID", None) == event["Unique-ID"]
            ):
                await session.fifo.put(event)
                semaphore.set()
                await cleanup()

        handlers["CHANNEL_EXECUTE_COMPLETE"] = channel_execute_complete_handler
        handlers["CHANNEL_HANGUP_COMPLETE"] = channel_hangup_complete_handler

        for key, value in handlers.items():
            self.on(key, partial(value, self))

        logger.debug(f"Register event handler for Application-UUID: {event_uuid}")

        return wait_for(semaphore, timeout=timeout)

    async def sendmsg(
        self,
        command: str,
        application: str,
        data: Optional[str] = None,
        lock: bool = False,
        uuid: Optional[str] = None,
        event_uuid: Optional[str] = None,
        block: bool = False,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[int] = None,
    ) -> ESLEvent:
        """
        Used to send commands from dialplan to session.

        Args:
            command: required
                Command to send to freeswitch. One of: execute, hangup, unicast, nomedia, xferext
            application: required
                Dialplan application to execute.
            data: optional
                Arguments to send to the application. If the command is 'hangup', this is the cause.
            lock: optional
                If true, lock the event.
            uuid: optional
                UUID of the call/session/channel to send the command.
            event_uuid: optional
                Adds a UUID to the execute command. In the corresponding events (CHANNEL_EXECUTE and
                CHANNEL_EXECUTE_COMPLETE), the UUID will be in the Application-UUID header.
            block: optional
                If true, wait for command completion before returning.
            headers: optional
                Additional headers to send with the command.
            timeout: optional
                Timeout for the command to complete. Just used if block is true.

        Raises:
            asyncio.TimeoutError: if the command does not complete within the timeout period.

        Returns:
            ESLEvent: The event received from FreeSWITCH after executing the command.
        """
        if uuid:
            cmd = f"sendmsg {uuid}"
        else:
            cmd = "sendmsg"

        cmd += f"\ncall-command: {command}"

        # Generate event_uuid if not provided and command is execute
        if command == "execute":
            cmd += f"\nexecute-app-name: {application}"
            if data:
                cmd += f"\nexecute-app-arg: {data}"

            event_uuid = event_uuid or str(uuid4())

            cmd += f"\nEvent-UUID: {event_uuid}"

        if lock:
            cmd += f"\nevent-lock: true"

        if command == "hangup":
            cmd += f"\nhangup-cause: {data}"

        if headers:
            for key, value in headers.items():
                cmd += f"\n{key}: {value}"

        logger.debug(f"Send command to freeswitch: '{cmd}'.")

        if block and command == "execute":
            logger.debug(
                f"Waiting for command completion with Application-UUID: {event_uuid}"
            )
            command_is_complete = await self._awaitable_complete_command(
                event_uuid, timeout
            )
            response = await self.send(cmd)
            logger.debug(
                f"Recived reponse of execute command with block: {pformat(response)}"
            )
            await command_is_complete.wait()
            return await self.fifo.get()

        return await self.send(cmd)

    async def log(
        self,
        level: Literal[
            "CONSOLE", "ALERT", "CRIT", "ERR", "WARNING", "NOTICE", "INFO", "DEBUG"
        ],
        message: str,
    ) -> ESLEvent:
        """Log a message to FreeSWITCH using dp tools log."""
        return await self.sendmsg("execute", "log", f"{level} {message}")

    async def answer(self) -> ESLEvent:
        """Answer the call associated with the session."""
        return await self.sendmsg("execute", "answer")

    async def park(self) -> ESLEvent:
        """Move session-associated call to park."""
        return await self.sendmsg("execute", "park")

    async def hangup(self, cause: str = "NORMAL_CLEARING") -> ESLEvent:
        """Hang up the call associated with the session."""
        return await self.sendmsg("execute", "hangup", cause)

    async def playback(
        self, path: str, block=True, timeout: Optional[int] = None
    ) -> ESLEvent:
        """Requests the freeswitch to play an audio."""
        return await self.sendmsg(
            "execute", "playback", path, block=block, timeout=timeout
        )

    async def say(
        self,
        text: str,
        module="en",
        lang: Optional[str] = None,
        kind: str = "NUMBER",
        method: str = "pronounced",
        gender: str = "FEMININE",
        block=True,
        timeout: Optional[int] = None,
    ) -> ESLEvent:
        """The say application will use the pre-recorded sound files to read or say things."""
        if lang:
            module += f":{lang}"

        arguments = f"{module} {kind} {method} {gender} {text}"
        logger.debug(f"Arguments used in say command: {arguments}")
        return await self.sendmsg(
            "execute", "say", arguments, block=block, timeout=timeout
        )

    async def play_and_get_digits(
        self,
        tries,
        timeout,
        terminators,
        file,
        minimal=0,
        maximum=128,
        block=True,
        regexp: Optional[str] = None,
        var_name: Optional[str] = None,
        invalid_file: Optional[str] = None,
        digit_timeout: Optional[int] = None,
        transfer_on_failure: Optional[str] = None,
        sendmsg_timeout: Optional[int] = None,
    ) -> ESLEvent:
        formatter = lambda value: "" if value is None else str(value)
        ordered_arguments = [
            minimal,
            maximum,
            tries,
            timeout,
            terminators,
            file,
            invalid_file,
            var_name,
            regexp,
            digit_timeout,
            transfer_on_failure,
        ]
        formated_ordered_arguments = map(formatter, ordered_arguments)
        arguments = " ".join(formated_ordered_arguments)
        logger.debug(f"Arguments used in play_and_get_digits command: {arguments}")

        return await self.sendmsg(
            "execute",
            "play_and_get_digits",
            arguments,
            block=block,
            timeout=sendmsg_timeout,
        )


class Outbound:
    """
    Outbound class
    -------------

    Given a valid set of information, start an ESL server that processes calls.

    Attributes:
    - host: required
        IP address the server should listen on.
    - port: required
        Network port the server should listen on.
    - handler: required
        Function that will take a session as an argument and will actually process the call.
    - events: optional
        If true, ask freeswitch to send us all events associated with the session.
    - linger: optional
        If true, asks that the events associated with the session come even after the call hangup.
    """

    def __init__(
        self,
        handler: Union[Callable, Coroutine],
        host: str = "127.0.0.1",
        port: int = 9000,
        events: bool = True,
        linger: bool = True,
    ) -> None:
        self.host = host
        self.port = port
        self.app = handler
        self.myevents = events
        self.linger = linger
        self.server = None

    async def start(self, block: bool = True) -> None:
        """Start the application server."""
        handler = partial(self.handler, self)
        self.server = await start_server(
            handler, self.host, self.port, family=socket.AF_INET
        )
        address = f"{self.host}:{self.port}"
        logger.info(f"Start application server and listen on '{address}'.")
        if block:
            await self.server.serve_forever()
        else:
            await self.server.start_serving()

    async def stop(self) -> None:
        """Terminate the application server."""
        if self.server:
            logger.debug("Shutdown application server.")
            self.server.close()
            await self.server.wait_closed()

    @staticmethod
    async def handler(
        server: Outbound, reader: StreamReader, writer: StreamWriter
    ) -> None:
        """Method used to process new connections."""
        with tracer.start_as_current_span(
            "outbound_handle_connection",
            attributes={
                "net.peer.name": server.host,
                "net.peer.port": server.port,
            },
        ):
            
            try:
                active_connections_counter.add(1, attributes={"type": "outbound"})
            except Exception:
                pass

            try:
                async with Session(reader, writer) as session:
                    logger.debug("Send command to start handle a call")
                    session.context = await session.send("connect")

                    if server.myevents:
                        logger.debug("Send command to receive all call events")
                        await session.send("myevents")

                    if server.linger:
                        logger.debug("Send linger command to freeswitch")
                        await session.send("linger")
                        session.is_lingering = True

                    logger.debug("Start server session handler")
                    await server.app(session)
            finally:
                try:
                    active_connections_counter.add(-1, attributes={"type": "outbound"})
                except Exception:
                    pass



import logging
import os
import json
import sys
from datetime import datetime, timezone
from rich.logging import RichHandler

TRACE_LEVEL_NUM = 5

logging.addLevelName(TRACE_LEVEL_NUM, "TRACE")


def trace(self, message, *args, **kws):
    if self.isEnabledFor(TRACE_LEVEL_NUM):
        self._log(TRACE_LEVEL_NUM, message, args, **kws)


logging.Logger.trace = trace


def get_log_level() -> int:
    """
    Get log level from environment variable or return default (INFO).

    Valid values for LOG_LEVEL are:
     - TRACE
     - DEBUG
     - INFO
     - WARNING
     - ERROR
     - CRITICAL

    Returns logging level constant.
    """
    level_map = {
        "TRACE": TRACE_LEVEL_NUM,
        "DEBUG": logging.DEBUG,
        "INFO": logging.INFO,
        "WARNING": logging.WARNING,
        "ERROR": logging.ERROR,
        "CRITICAL": logging.CRITICAL,
    }

    env_level = os.getenv("GENESIS_LOG_LEVEL", "INFO").upper()
    return level_map.get(env_level, logging.INFO)


from opentelemetry import trace


class CorrelationIdFilter(logging.Filter):
    """
    Log filter to inject the current OpenTelemetry trace and span IDs into the log record.
    """

    def filter(self, record):
        try:
            span = trace.get_current_span()
            if span == trace.INVALID_SPAN:
                return True
            
            ctx = span.get_span_context()
            trace_id = trace.format_trace_id(ctx.trace_id)
            span_id = trace.format_span_id(ctx.span_id)
            
            record.otelTraceID = trace_id
            record.otelSpanID = span_id

            if isinstance(record.msg, str):
                record.msg = f"{record.msg} (trace_id={trace_id} span_id={span_id})"
        except Exception:
            pass
            
        return True


class JSONFormatter(logging.Formatter):
    """
    Formatter that outputs JSON strings.
    """

    def format(self, record):
        log_record = {
            "timestamp": datetime.fromtimestamp(record.created, tz=timezone.utc).isoformat(),
            "level": record.levelname,
            "message": record.getMessage(),
            "logger": record.name,
        }
        
        if record.exc_info:
            log_record["exception"] = self.formatException(record.exc_info)
            
        if hasattr(record, "otelTraceID") and record.otelTraceID != "0":
            log_record["trace_id"] = record.otelTraceID
        if hasattr(record, "otelSpanID") and record.otelSpanID != "0":
            log_record["span_id"] = record.otelSpanID
            
        return json.dumps(log_record)


def reconfigure_logger(use_json: bool = False) -> None:
    """
    Reconfigure the logger to use JSON format or Rich format.
    
    Args:
        use_json: If True, use JSONFormatter. If False, use RichHandler.
    """
    logger = logging.getLogger(__name__)
    
    for handler in logger.handlers[:]:
        logger.removeHandler(handler)
        
    if use_json:
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(JSONFormatter())
    else:
        handler = RichHandler(
            rich_tracebacks=True,
            tracebacks_show_locals=True,
            markup=True,
            show_path=False,
            show_time=True,
            omit_repeated_times=False,
        )
        handler.setFormatter(logging.Formatter("%(message)s"))
    
    
    handler.addFilter(CorrelationIdFilter())
    
    logger.addHandler(handler)
    logger.debug(f"Logger reconfigured (json={use_json})")


def setup_logger(name: str = __name__) -> logging.Logger:
    """Configure a logger with rich handler and conventional formatting.

    Args:
        name: The name for the logger instance

    Returns:
        A configured logger instance
    """
    logger = logging.getLogger(name)

    if logger.handlers:
        return logger

    handler = RichHandler(
        rich_tracebacks=True,
        tracebacks_show_locals=True,
        markup=True,
        show_path=False,
        show_time=True,
        omit_repeated_times=False,
    )
    
    
    handler.addFilter(CorrelationIdFilter())

    formatter = logging.Formatter("%(message)s")
    handler.setFormatter(formatter)

    
    log_level = get_log_level()
    logger.setLevel(log_level)
    logger.addHandler(handler)
    logger.propagate = False

    logger.debug(f"Logger initialized with level: {logging.getLevelName(log_level)}")
    return logger


logger = setup_logger(__name__)



"""
CLI module for Genesis.
------------------------

This module contains the CLI commands for Genesis.
"""

import importlib.metadata
import os
from typing import Annotated, Union

import typer
from rich import print
from opentelemetry import metrics
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.resources import Resource
from opentelemetry.exporter.prometheus import PrometheusMetricReader
from prometheus_client import start_http_server

from genesis.cli.consumer import consumer
from genesis.cli.outbound import outbound
from genesis.logger import reconfigure_logger


app = typer.Typer(rich_markup_mode="rich")
app.add_typer(consumer, name="consumer", short_help="Run you ESL events consumer.")
app.add_typer(outbound, name="outbound", short_help="Run you outbound services.")


def version(show: bool) -> None:
    """Show the version and exit."""
    if show:
        version = importlib.metadata.version("genesis")
        logger.info(f"Genesis version: {version}")
        raise typer.Exit()


@app.callback()
def callback(
    version: Annotated[
        Union[bool, None],
        typer.Option("--version", help="Show the version and exit.", callback=version),
    ] = None,
    json: Annotated[
        bool,
        typer.Option("--json", help="Output logs in JSON format."),
    ] = False,
) -> None:
    reconfigure_logger(json)
    
    
    try:
        metrics_port = int(os.getenv("GENESIS_METRICS_PORT", "8000"))
        start_http_server(metrics_port)
        
        metric_reader = PrometheusMetricReader()
        provider = MeterProvider(
            resource=Resource.create({"service.name": "genesis"}),
            metric_readers=[metric_reader],
        )
        metrics.set_meter_provider(provider)
        
        logger.info(f"Prometheus metrics server started on port {metrics_port}")
    except Exception as e:
        
        logger.warning(f"Failed to start Prometheus metrics server on port {metrics_port}: {e}")
    
    """
    Genesis - [blue]FreeSWITCH Event Socket protocol[/blue] implementation with [bold]asyncio[/bold].

    Run yours freeswitch apps without any external dependencies.

    ℹ️ Read more in the docs: [link]https://otoru.github.io/Genesis/[/link].
    """



"""
Genesis Protocol
----------------

Here we will group what is common to the ESL client for inbound and outbound connections.
"""

from asyncio import (
    iscoroutinefunction,
    StreamWriter,
    StreamReader,
    create_task,
    to_thread,
    Event,
    Queue,
    Task,
)

from typing import List, Dict, Optional, Callable, Coroutine, Any, Union
from abc import ABC
import logging

import time
from opentelemetry import trace, metrics
from rich.console import Console

from genesis.exceptions import ConnectionError, UnconnectedError
from genesis.logger import logger, TRACE_LEVEL_NUM
from genesis.parser import ESLEvent, parse_headers

tracer = trace.get_tracer(__name__)
meter = metrics.get_meter(__name__)

commands_sent_counter = meter.create_counter(
    "genesis.commands.sent",
    description="Number of ESL commands sent",
    unit="1",
)
events_received_counter = meter.create_counter(
    "genesis.events.received",
    description="Number of ESL events received",
    unit="1",
)
command_duration_histogram = meter.create_histogram(
    "genesis.commands.duration",
    description="Duration of ESL commands execution",
    unit="s",
)
command_errors_counter = meter.create_counter(
    "genesis.commands.errors",
    description="Number of failed ESL commands",
    unit="1",
)


class Protocol(ABC):
    def __init__(self):
        self.events = Queue()
        self.commands = Queue()
        self.is_connected = False
        self.is_lingering = False
        self.authentication_event = Event()
        self.producer: Optional[Task] = None
        self.consumer: Optional[Task] = None
        self.reader: Optional[StreamReader] = None
        self.writer: Optional[StreamWriter] = None
        self.handlers: Dict[
            str,
            List[
                Union[
                    Callable[[ESLEvent], None],
                    Callable[[ESLEvent], Coroutine[Any, Any, None]],
                ]
            ],
        ] = {}

    async def start(self) -> None:
        """Initiates a connection to a freeswitch."""
        self.is_connected = True

        logger.debug("Create tasks to work with ESL events.")
        self.producer = create_task(self.handler())
        self.consumer = create_task(self.consume())

    async def stop(self) -> None:
        """Terminates connection to a freeswitch."""
        if self.writer and not self.writer.is_closing():
            logger.debug("Closer stream writer.")
            self.writer.close()

        self.is_connected = False

        if self.producer and not self.producer.cancelled():
            logger.debug("Cancel event producer task.")
            self.producer.cancel()

        if self.consumer and not self.consumer.cancelled():
            logger.debug("Cancel event consumer task.")
            self.consumer.cancel()

    async def handler(self) -> None:
        """Defines intelligence to treat received events."""
        while self.is_connected:
            request = None
            buffer = ""

            while self.is_connected:
                try:
                    content = await self.reader.readline()

                    if not content:
                        self.is_connected = False
                        break
                    
                    buffer += content.decode("utf-8")
                except Exception as e:
                    logger.error(f"Error reading from stream. {str(e)}")
                    self.is_connected = False
                    break

                if buffer[-2:] == "\n\n" or buffer[-4:] == "\r\n\r\n":
                    request = buffer
                    logger.trace(f"Complete message received: {repr(request)}")
                    break

            if not request or not self.is_connected:
                break

            event = parse_headers(request)

            if "Content-Length" in event:
                # Get the total length from the first Content-Length header
                length = int(event["Content-Length"].split("\n")[0])
                logger.trace(f"Total content length: {length} bytes")

                # Read the complete data
                data = await self.reader.readexactly(length)
                logger.trace(f"Received complete data: {data}")
                complete_content = data.decode("utf-8")
                contentType = event.get("Content-Type", None)

                if contentType:
                    logger.trace(f"Check content type of event: {event}")

                    if contentType not in [
                        "api/response",
                        "text/rude-rejection",
                        "log/data",
                    ]:
                        # Try to split headers and body
                        if "\n\n" in complete_content:
                            headers_part, body = complete_content.split("\n\n", 1)

                            # Here we check for multiple events in one message (can happen if event-lock is set)
                            event_parts = []

                            if "event-lock: true" in headers_part.lower():
                                # Split the string on "Event-Name: "
                                parts = headers_part.split("\nEvent-Name: ")

                                if len(parts) > 1:
                                    event_parts = [parts[0]]

                                    for part in parts[1:]:
                                        event_parts.append(f"Event-Name: {part}")

                                    logger.debug(
                                        f"Split locked event into {len(event_parts)} separate events"
                                    )
                            else:
                                event_parts = [headers_part]

                            # Process each event part
                            for idx, event_str in enumerate(event_parts):
                                if idx == 0:
                                    # First event is the original event
                                    additional_headers = parse_headers(event_str)
                                    event.update(additional_headers)
                                    event.body = body
                                    await self.events.put(event)
                                else:
                                    # More events are new events
                                    new_event = parse_headers(event_str)
                                    # Copy some headers from the original event
                                    for key in ["Content-Length", "Content-Type"]:
                                        if key in event:
                                            new_event[key] = event[key]
                                    new_event.body = body
                                    await self.events.put(new_event)
                            continue  # Skip the final event.put

                        else:
                            # If no clear header/body separation, treat everything as body
                            event.body = complete_content
                    else:
                        event.body = complete_content
                else:
                    event.body = complete_content

            await self.events.put(event)

    async def consume(self) -> None:
        """Arm all event processors."""
        while self.is_connected:
            event = await self.events.get()

            try:
                # OTel Tracing
                attributes = {}
                for key, value in event.items():
                    if key == "Event-Name":
                        attr_name = "event.name"
                    elif key == "Unique-ID":
                        attr_name = "event.uuid"
                    elif key == "Content-Type":
                        attr_name = "event.content_type"
                    else:
                        slug = key.lower().replace("-", "_")
                        attr_name = f"event.header.{slug}"
                    
                    if isinstance(value, (str, int, float, bool, list, tuple)):
                        attributes[attr_name] = value

                try:
                    with tracer.start_as_current_span("process_event", attributes=attributes) as span:
                        for header_name, header_value in event.items():
                            # Encode headers if necessary or ensure string
                            attribute_key = header_name.lower().replace("-", "_")
                            span.set_attribute(f"event.header.{attribute_key}", str(header_value))
                        
                        
                        event_name = event.get("Event-Name", "UNKNOWN")
                        content_type = event.get("Content-Type", "UNKNOWN")
                        metric_attributes = {
                            "event_name": event_name,
                            "content_type": content_type,
                        }
                        
                        
                        if "Event-Subclass" in event:
                            metric_attributes["event_subclass"] = event["Event-Subclass"]
                        if "Call-Direction" in event:
                            metric_attributes["direction"] = event["Call-Direction"]
                        if "Channel-State" in event:
                            metric_attributes["channel_state"] = event["Channel-State"]
                        if "Answer-State" in event:
                            metric_attributes["answer_state"] = event["Answer-State"]
                        if "Hangup-Cause" in event:
                            metric_attributes["hangup_cause"] = event["Hangup-Cause"]
                except Exception:
                    # OTel not initialized or error in tracing - continue without tracing
                    event_name = event.get("Event-Name", "UNKNOWN")
                    content_type = event.get("Content-Type", "UNKNOWN")
                    metric_attributes = {
                        "event_name": event_name,
                        "content_type": content_type,
                    }
                    
                try:
                    events_received_counter.add(1, attributes=metric_attributes)
                except Exception:
                    pass

                # LOGGING
                try:
                    if logger.isEnabledFor(TRACE_LEVEL_NUM):
                        logger.trace(f"Received an event: '{event}'.")

                    else:
                        if logger.isEnabledFor(logging.DEBUG):
                            name = event.get("Event-Name", None)
                            uuid = event.get("Unique-ID", None)

                            if uuid:
                                logger.debug(
                                    f"Received an event: '{name}' for call '{uuid}'. "
                                )

                                if name == "CHANNEL_EXECUTE_COMPLETE":
                                    application = event.get("Application")
                                    response = event.get("Application-Response")

                                    logger.debug(
                                        f"Application: '{application}' - Response: '{response}'."
                                    )

                            else:
                                if name:
                                    logger.debug(f"Received an event: '{name}'.")

                                elif "Content-Type" in event and event["Content-Type"] in [
                                    "command/reply",
                                    "auth/request",
                                ]:
                                    reply = event.get("Reply-Text", None)

                                    if reply and event["Content-Type"] == "command/reply":
                                        logger.debug(
                                            f"Received an command reply: '{reply}'."
                                        )

                                    if reply and event["Content-Type"] == "auth/request":
                                        logger.debug(
                                            f"Received an authentication reply: '{event}'."
                                        )

                except Exception as e:
                    logger.error(f"Error logging event: {str(e)} - Event: {event}")

                if "Content-Type" in event and event["Content-Type"] == "auth/request":
                    self.authentication_event.set()

                elif "Content-Type" in event and event["Content-Type"] == "command/reply":
                     await self.commands.put(event)

                elif "Content-Type" in event and event["Content-Type"] == "api/response":
                     await self.commands.put(event)

                elif "Content-Type" in event and event["Content-Type"] in [
                    "text/rude-rejection",
                    "text/disconnect-notice",
                ]:
                    if not (
                        "Content-Disposition" in event
                        and event["Content-Disposition"] == "linger"
                    ):
                        await self.stop()

                identifier = event.get("Event-Name", None)

                if identifier == "CUSTOM":
                    name = event.get("Event-Subclass", None)
                else:
                    name = identifier

                if name:
                    logger.trace(f"Get all handlers for '{name}'.")
                    specific = self.handlers.get(name, [])
                    generic = self.handlers.get("*", list())
                    handlers = specific + generic

                    if handlers:
                        for handler in handlers:
                            if iscoroutinefunction(handler):
                                create_task(handler(event))
                            else:
                                create_task(to_thread(handler, event))

            except Exception as outer_e:
                logger.error(f"Error in consumer loop: {outer_e}", exc_info=True)
                # Continue loop to avoid freezing connection

    def on(
        self,
        key: str,
        handler: Union[
            Callable[[ESLEvent], None], Callable[[ESLEvent], Coroutine[Any, Any, None]]
        ],
    ) -> None:
        """Associate a handler with an event key."""
        logger.debug(f"Register handler to '{key}' event.")
        self.handlers.setdefault(key, list()).append(handler)

    def remove(
        self,
        key: str,
        handler: Union[
            Callable[[ESLEvent], None], Callable[[ESLEvent], Coroutine[Any, Any, None]]
        ],
    ) -> None:
        """Removes the HANDLER from the list of handlers for the given event KEY name."""
        logger.debug(f"Remove handler to '{key}' event.")
        if key in self.handlers and handler in self.handlers[key]:
            self.handlers.setdefault(key, list()).remove(handler)

    async def send(self, cmd: str) -> ESLEvent:
        """Method used to send commands to or freeswitch."""
        if not self.is_connected:
            raise UnconnectedError()

        if self.writer.is_closing():
            raise ConnectionError()


        start_time = time.perf_counter()
        try:
            with tracer.start_as_current_span("send_command") as span:
                span.set_attribute("command.name", cmd)
                if "Event-UUID" in self.events:
                    span.set_attribute("command.uuid", self.events["Event-UUID"])

                logger.debug(f"Send command: {cmd}")
                try:
                    command_name = cmd.split()[0]
                    commands_sent_counter.add(1, attributes={"command": command_name})
                except Exception:
                    pass

                try:
                    self.writer.write(f"{cmd}\n\n".encode())
                    await self.writer.drain()
                    result = await self.commands.get()
                    
                    reply = result.get("Reply-Text", "")
                    if reply.startswith("-ERR"):
                        try:
                            command_errors_counter.add(1, attributes={"command": command_name, "error": "protocol_error"})
                        except Exception:
                            pass
                    
                    reply_text = result.get("Reply-Text")
                    if reply_text:
                        span.set_attribute("command.reply", reply_text)

                    return result
                except Exception as e:
                    try:
                        command_errors_counter.add(1, attributes={"command": command_name, "error": type(e).__name__})
                    except Exception:
                        pass
                    raise
                finally:
                    try:
                        duration = time.perf_counter() - start_time
                        command_duration_histogram.record(duration, attributes={"command": command_name})
                    except Exception:
                        pass
        except Exception:
            # OTel not initialized - run without tracing
            logger.debug(f"Send command: {cmd}")
            try:
                command_name = cmd.split()[0]
                commands_sent_counter.add(1, attributes={"command": command_name})
            except Exception:
                pass

            try:
                self.writer.write(f"{cmd}\n\n".encode())
                await self.writer.drain()
                result = await self.commands.get()
                
                reply = result.get("Reply-Text", "")
                if reply.startswith("-ERR"):
                    try:
                        command_errors_counter.add(1, attributes={"command": command_name, "error": "protocol_error"})
                    except Exception:
                        pass
                
                return result
            except Exception as e:
                try:
                    command_errors_counter.add(1, attributes={"command": command_name, "error": type(e).__name__})
                except Exception:
                    pass
                raise
            finally:
                try:
                    duration = time.perf_counter() - start_time
                    command_duration_histogram.record(duration, attributes={"command": command_name})
                except Exception:
                    pass



import logging


def complete_log_levels(incomplete: str):
    """Autocompletion for log levels."""
    levels = [item.lower() for item in logging.getLevelNamesMapping().keys()]

    for item in levels:
        if item.startswith(incomplete):
            yield item



"""
Genesis exceptions
------------------

Grouping of all errors that can occur in genesis.
"""


class ConnectionError(ConnectionError):
    """exception created to group all connection errors."""

    ...


class ConnectionTimeoutError(ConnectionError):
    """Occurs when the connection does not occur within the established timeout."""

    ...


class SessionGoneAway(Exception):
    """Occurs when the session has already received a hangup."""

    ...


class AuthenticationError(ValueError):
    """It happens when we have a problem during authentication."""

    ...


class UnconnectedError(Exception):
    """It happens when we try to send dice to a server once we connect before."""

    ...



import importlib.metadata

from .consumer import Consumer, filtrate
from .outbound import Outbound, Session, ESLEvent
from .inbound import Inbound

__all__ = ["Inbound", "Consumer", "filtrate", "Outbound", "Session", "ESLEvent"]
__version__ = importlib.metadata.version("genesis")



import sys
import importlib
from typing import Union
from pathlib import Path



from genesis.logger import logger
from genesis.cli.exceptions import CLIExcpetion


def get_app_name(
    cls: object, module_import_str: str, app_name: Union[str, None] = None
) -> str:
    """Get the app name from the module."""
    try:
        mod = importlib.import_module(module_import_str)
    except (ImportError, ValueError):
        raise CLIExcpetion(
            "Ensure all the package directories have an [blue]__init__.py[/blue] file"
        )
    object_names = dir(mod)
    object_names_set = set(object_names)

    if app_name:
        if app_name not in object_names_set:
            raise CLIExcpetion(
                f"Could not find app name {app_name} in {module_import_str}"
            )

        app = getattr(mod, app_name)

        if not isinstance(app, cls):
            raise CLIExcpetion(
                f"The app name {app_name} in {module_import_str} doesn't seem to be a {cls.__name__} app"
            )

        return app_name

    for name in object_names:
        obj = getattr(mod, name)
        if isinstance(obj, cls):
            return name

    raise CLIExcpetion(f"Could not find {cls.__name__} app in module, try using --app")


def get_import_string(
    cls: object, path: Union[Path, None] = None, app_name: Union[str, None] = None
) -> str:
    """Get the import string for the given module path."""
    if not path:
        raise CLIExcpetion("A path to a Python file or package directory is required.")

    logger.info(f"Resolved absolute path {path.resolve()}")

    if not path.exists():
        raise CLIExcpetion(f"The path '{path}' does not exist.")

    absolute_path = path.resolve()
    module_path = absolute_path

    if absolute_path.is_file() and absolute_path.stem == "__init__":
        module_path = absolute_path.parent

    module_paths = [module_path]
    extra_sys_path = module_path.parent

    for parent in module_path.parents:
        init_path = parent / "__init__.py"

        if init_path.is_file():
            module_paths.insert(0, parent)
            extra_sys_path = parent.parent

        else:
            break

    logger.info(f"Module structure: {' -> '.join(p.name for p in module_paths)}")

    module_import_str = ".".join(p.stem for p in module_paths)

    logger.info(f"Importing module [green]{module_import_str}[/green]")

    extra_sys_path = extra_sys_path.resolve()

    sys.path.insert(0, str(extra_sys_path))

    use_app_name = get_app_name(cls, module_import_str, app_name=app_name)

    logger.info(f"Importable app: from {module_import_str} import {use_app_name}")

    import_string = f"{module_import_str}:{use_app_name}"
    logger.info(f"Using import string [b green]{import_string}[/b green]")

    return import_string



from typing import Annotated, Union
from pathlib import Path
import importlib
import asyncio
import logging

import typer


from genesis.cli import watcher
from genesis.logger import logger
from genesis.outbound import Outbound
from genesis.cli.exceptions import CLIExcpetion
from genesis.cli.utils import complete_log_levels
from genesis.cli.discover import get_import_string


outbound = typer.Typer(rich_markup_mode="rich")


@outbound.command()
def dev(
    path: Annotated[
        Path,
        typer.Argument(
            help="A path to a Python file or package directory.", metavar="PATH"
        ),
    ],
    *,
    host: Annotated[
        str,
        typer.Option(help="The host to serve on.", envvar="ESL_APP_HOST"),
    ] = "127.0.0.1",
    port: Annotated[
        int,
        typer.Option(help="The port to serve on.", envvar="ESL_APP_PORT"),
    ] = 9000,
    app: Annotated[
        Union[str, None],
        typer.Option(
            help="Variable that contains the [bold]Outbound[/bold] app in the imported module or package.",
            envvar="ESL_APP_NAME",
        ),
    ] = None,
    loglevel: Annotated[
        str,
        typer.Option(
            help="The log level to use.",
            envvar="ESL_LOG_LEVEL",
            show_default=True,
            case_sensitive=False,
            autocompletion=complete_log_levels,
        ),
    ] = "info",
):
    """
    Run a [bold]Outbound[/bold] genesis app in [yellow]development[/yellow] mode. 🧪

    It automatically detects the Python module or package that needs to be imported based on the file or directory path passed.

    It detects the [bold]Outbound[/bold] app object to use based on the app name passed.
    By default it looks in the module or package for an object named [blue]app[/blue].
    Otherwise, it uses the first [bold]Outbound[/bold] app found in the imported module or package.
    """
    _run(path=path, host=host, port=port, app=app, reload=True, loglevel=loglevel)


@outbound.command()
def run(
    path: Annotated[
        Path,
        typer.Argument(
            help="A path to a Python file or package directory.", metavar="PATH"
        ),
    ],
    *,
    host: Annotated[
        str,
        typer.Option(help="The host to serve on.", envvar="ESL_APP_HOST"),
    ] = "127.0.0.1",
    port: Annotated[
        int,
        typer.Option(help="The port to serve on.", envvar="ESL_APP_PORT"),
    ] = 9000,
    app: Annotated[
        Union[str, None],
        typer.Option(
            help="Variable that contains the [bold]Outbound[/bold] app in the imported module or package.",
            envvar="ESL_APP_NAME",
        ),
    ] = None,
    loglevel: Annotated[
        str,
        typer.Option(
            help="The log level to use.",
            envvar="ESL_LOG_LEVEL",
            show_default=True,
            case_sensitive=False,
            autocompletion=complete_log_levels,
        ),
    ] = "info",
):
    """
    Run a [bold]Outbound[/bold] genesis app in [green]production[/green] mode. 🚀

    It automatically detects the Python module or package that needs to be imported based on the file or directory path passed.

    It detects the [bold]Outbound[/bold] app object to use based on the app name passed.
    By default it looks in the module or package for an object named [blue]app[/blue].
    Otherwise, it uses the first [bold]Outbound[/bold] app found in the imported module or package.
    """
    _run(path=path, host=host, port=port, app=app, reload=False, loglevel=loglevel)


async def _run_with_reload(app: Outbound, path: Path) -> None:
    loop = asyncio.get_running_loop()
    queue = asyncio.Queue()

    async def consume(queue: asyncio.Queue) -> None:
        await app.start(block=False)
        async for event in watcher.EventIterator(queue):
            if event:
                logger.info(f"File changed: {event.src_path}")
                await app.stop()
                logger.info("App stopped, restarting...")
                await app.start(block=False)

    observer = watcher.factory(path, queue, loop)

    observer.start()
    asyncio.run_coroutine_threadsafe(queue.put(None), loop)

    try:
        await consume(queue)
        while True:
            await asyncio.sleep(1)
    except KeyboardInterrupt:
        observer.stop()

    observer.join()


def _run(
    path: Union[Path, None] = None,
    host: str = "127.0.0.1",
    port: int = 9000,
    reload: bool = True,
    app: Union[str, None] = None,
    loglevel: str = "info",
) -> None:
    try:
        import_string = get_import_string(Outbound, path=path, app_name=app)

        logger.info(f"Outbound server started - Host: {host}, Port: {port}")

        module_str, attr_str = import_string.split(":")
        module = importlib.import_module(module_str)
        app: Outbound = getattr(module, attr_str)

        app.host = host
        app.port = port

        logger.info(f"Setting log level to [bold]{loglevel.upper()}[/bold]")
        levels = logging.getLevelNamesMapping()
        logger.setLevel(levels.get(loglevel.upper(), logging.INFO))

        if reload:
            asyncio.run(_run_with_reload(app, path))
        else:
            asyncio.run(app.start())

    except CLIExcpetion as e:
        logger.error(e)
        raise typer.Exit(1)
